import re


def to_re(m):
    if m.group(0) == '*':
        return '.*'
    elif m.group(0) == '?':
        return '.'
    else:
        return re.escape(m.group(0))


class TagPattern(object):
    def __init__(self, raw):
        self.raw = raw

        as_re = re.sub(r'\*|\?|[^*?]*', to_re, raw) + '$'
        self.compiled = re.compile(as_re)

    def __eq__(self, other):
        return self.raw == other.raw

    def __hash__(self):
        return hash(self.raw)

    def matches(self, tag):
        return self.compiled.match(tag) is not None

    def __repr__(self):
        return 'TagPattern({!r})'.format(self.raw)


if __name__ == '__main__':
    tp = TagPattern('latest')
    assert not tp.matches('late')
    assert tp.matches('latest')
    tp = TagPattern('?atest')
    assert tp.matches('latest')
    assert not tp.matches('llatest')
    tp = TagPattern('?ate*')
    assert tp.matches('latest')
    assert tp.matches('latester')
    assert not tp.matches('')

    s = {TagPattern(x) for x in ['latest', 'lat*', 'latest']}
    assert len(s) == 2



import click
import logging

from regindexer.indexer import Indexer


@click.group()
@click.option('-c', '--config', metavar='CONFIG_FILE',
              help='Path to config file',
              default='/etc/regindexer/config.yaml')
@click.option('-v', '--verbose', is_flag=True,
              help='Show verbose debugging output')
@click.pass_context
def cli(ctx, config, verbose):
    ctx.obj = {
        'config': config
    }

    if verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)


@cli.command(name="index")
@click.pass_context
def index(ctx):
    """Rebuild indexes"""
    indexer = Indexer(config=ctx.obj['config'])
    indexer.index()



import yaml

from regindexer.tag_pattern import TagPattern


class IndexConfig(object):
    def __init__(self, name, attrs):
        self.name = name
        self.output = attrs['output']
        self.registry = attrs['registry']
        self.registry_public = attrs.get('registry_public', self.registry)
        self.tags = [TagPattern(x) for x in attrs['tags']]
        self.required_annotations = attrs.get('required_annotations', [])
        self.architectures = attrs.get('architectures', None)
        self.extract_icons = attrs.get('extract_icons', False)

    def __repr__(self):
        return 'IndexConfig(%r)' % self.__dict__


class Config(object):
    def __init__(self, path):
        self.indexes = []
        with open(path, 'r') as f:
            yml = yaml.safe_load(f)
            self.icons_dir = yml.get('icons_dir', None)
            self.icons_uri = yml.get('icons_uri', None)
            if not self.icons_uri.endswith('/'):
                self.icons_uri = self.icons_uri + '/'
            indexes = yml.get('indexes')
            if indexes:
                for name, attrs in indexes.items():
                    self.indexes.append(IndexConfig(name, attrs))



import logging

from fedmsg.consumers import FedmsgConsumer
import moksha.hub.reactor
import six.moves.queue as queue

from regindexer.indexer import Indexer

log = logging.getLogger('regindexer.consumer')


class Consumer(FedmsgConsumer):
    config_key = 'regindexer.consumer.enabled'

    def __init__(self, hub):
        super(Consumer, self).__init__(hub)

        prefix = hub.config.get('topic_prefix')
        env = hub.config.get('environment')
        self.topic = [
            prefix + '.' + env + '.bodhi.mashtask.complete',
        ]

        # Need to do this explicitly to avoid triggering the Initialization task
        if not hub.config.get(self.config_key, False):
            return

        self.index_queue = queue.Queue()

        config_file = hub.config.get('regindexer.config_file',
                                     '/etc/regindexer/config.yaml')
        self.indexer = Indexer(config_file)

        moksha.hub.reactor.reactor.callInThread(self._do_index_work)
        self.index_queue.put('Initialization')

    def consume(self, raw_msg):
        topic = raw_msg['topic']
        msg = raw_msg['body']['msg']

        if msg.get('ctype') in ('container', 'flatpak'):
            log.info("Got %s message for ctype=%s", topic, msg['ctype'])
            # trigger rebuilding the index
            self.index_queue.put('Bodhi mash, ctype="%s"' % msg['ctype'])

    def _do_index_work(self):
        while True:
            reason = self.index_queue.get()
            log.info("Rebuilding index for: %s", reason)
            self.indexer.index()



config = {
    'regindexer.consumer.enabled': False,
    'regindexer.config_file': '/etc/regindexer/config.yaml',
    'logging': {
        'loggers': {
            'regindexer': {
                "level": "INFO",
                "propagate": False,
                "handlers": ["console"],
            },
            'regindexer.consumer': {
                "level": "INFO",
                "propagate": False,
                "handlers": ["console"],
            }
        }
    }
}



from regindexer.config import Config

import base64
import hashlib
import logging
import json
import requests
import os
import sys
from urllib.parse import urljoin

log = logging.getLogger('regindexer')


MEDIA_TYPE_MANIFEST_V2 = 'application/vnd.docker.distribution.manifest.v2+json'
MEDIA_TYPE_LIST_V2 = 'application/vnd.docker.distribution.manifest.list.v2+json'
MEDIA_TYPE_OCI = 'application/vnd.oci.image.manifest.v1+json'
MEDIA_TYPE_OCI_INDEX = 'application/vnd.oci.image.index.v1+json'

MANIFEST_HEADERS = {
    'Accept': ', '.join((
        MEDIA_TYPE_MANIFEST_V2,
        MEDIA_TYPE_LIST_V2,
        MEDIA_TYPE_OCI,
        MEDIA_TYPE_OCI_INDEX,
    ))
}

DATA_URI_PREFIX = 'data:image/png;base64,'


class DownloadInfo(object):
    def __init__(self, digest, content_type, json):
        self.digest = digest
        self.content_type = content_type
        self.json = json

    def __repr__(self):
        return 'DownloadInfo(%r)' % self.__dict__


class RegistryQuery(object):
    def __init__(self, registry, verbose=False):
        self.registry = registry
        self.verbose = verbose
        self.session = requests.Session()

    def _get(self, relative_url, headers=None):
        return self.session.get(self.registry + relative_url,
                                headers=headers)

    def get_manifest(self, name, ref):
        log.info("Querying %s:%s", name, ref)

        url = '/v2/{}/manifests/{}'.format(name, ref)
        response = self._get(url, headers=MANIFEST_HEADERS)
        if response.status_code != 200:
            log.warning("Could not download %s:%s", name, ref)
            return None

        return DownloadInfo(response.headers['Docker-Content-Digest'],
                            response.headers['Content-Type'],
                            response.json())

    def get_blob(self, name, digest):
        url = '/v2/{}/blobs/{}'.format(name, digest)
        response = self._get(url)
        if response.status_code != 200:
            return None

        return DownloadInfo(response.headers['Docker-Content-Digest'],
                            response.headers['Content-Type'],
                            response.json())

    def iterate_images(self, tag_patterns):
        repo_names = self._get('/v2/_catalog').json()['repositories']

        for name in repo_names:
            url = '/v2/{}/tags/list'.format(name)
            tags = self._get(url).json()['tags']
            matches = set()

            for tag in tags:
                for pat in tag_patterns:
                    if pat.matches(tag):
                        matches.add(tag)

            manifests = {}
            tags = {}
            for tag in matches:
                manifest_info = self.get_manifest(name, tag)
                if not manifest_info:
                    continue

                manifests[manifest_info.digest] = manifest_info
                tags.setdefault(manifest_info.digest, []).append(tag)

            for digest in manifests:
                yield name, manifests[digest], tags[digest]


class IconStore(object):
    def __init__(self, icons_dir, icons_uri):
        self.icons_dir = icons_dir
        self.icons_uri = icons_uri

        if not os.path.exists(icons_dir):
            # Create only one level
            os.mkdir(icons_dir)

        self.old_icons = {}
        for subdir in os.listdir(icons_dir):
            for filename in os.listdir(os.path.join(icons_dir, subdir)):
                self.old_icons[(subdir, filename)] = True

        self.icons = {}

    def store(self, uri):
        if not uri.startswith(DATA_URI_PREFIX):
            return None

        decoded = base64.b64decode(uri[len(DATA_URI_PREFIX):])

        h = hashlib.sha256()
        h.update(decoded)
        digest = h.hexdigest()
        subdir = digest[:2]
        filename = digest[2:] + '.png'

        key = (subdir, filename)
        if key in self.icons:
            pass
        elif key in self.old_icons:
            self.icons[key] = True
        else:
            if not os.path.exists(os.path.join(self.icons_dir, subdir)):
                os.mkdir(os.path.join(self.icons_dir, subdir))
            fullpath = os.path.join(self.icons_dir, subdir, filename)
            log.info("Storing icon: %s", fullpath)
            with open(os.path.join(self.icons_dir, subdir, filename), 'wb') as f:
                f.write(decoded)
            self.icons[key] = True

        return urljoin(self.icons_uri, subdir + '/' + filename)

    def clean(self):
        for key in self.old_icons:
            if key not in self.icons:
                subdir, filename = key
                fullpath = os.path.join(self.icons_dir, subdir, filename)
                os.unlink(fullpath)
                log.info("Removing icon: %s", fullpath)


class Index(object):
    def __init__(self, conf, icon_store=None):
        if conf.extract_icons and icon_store is None:
            raise RuntimeError("extract_icons is set, but no icons_dir is configured")

        self.config = conf
        self.icon_store = icon_store
        self.repos = {}

    def extract_icon(self, annotations, key):
        if not self.config.extract_icons:
            return

        value = annotations.get(key)
        if value is None:
            return

        uri = self.icon_store.store(value)
        if uri is not None:
            annotations[key] = uri

    def make_image(self, query, name, manifest_info, tags=None):
        config_info = query.get_blob(name, manifest_info.json['config']['digest'])
        if not config_info:
            log.warning("Failed to download config json")
            return None

        arch = config_info.json['architecture']
        os = config_info.json['os']

        if self.config.architectures:
            if arch not in self.config.architectures:
                return None

        if manifest_info.content_type == MEDIA_TYPE_OCI:
            annotations = manifest_info.json['annotations']
        else:
            annotations = {}

        for annotation in self.config.required_annotations:
            if annotation not in annotations:
                return None

        self.extract_icon(annotations, 'org.freedesktop.appstream.icon-64')
        self.extract_icon(annotations, 'org.freedesktop.appstream.icon-128')

        image = {
            'Digest': manifest_info.digest,
            'MediaType': manifest_info.content_type,
            'OS': os,
            'Architecture':  arch,
            'Labels': {},
        }

        image['Annotations'] = annotations

        labels = config_info.json['config'].get('Labels')
        if labels is None:
            labels = {}
        image['Labels'] = labels

        if tags:
            image['Tags'] = tags

        return image

    def make_list(self, query, name, list_info, tags=None):
        images = []
        for manifest in list_info.json['manifests']:
            manifest_info = query.get_manifest(name, manifest['digest'])

            if not manifest_info:
                log.warning("Failed to download manifest {}".format(manifest['digest']),
                            file=sys.stderr)
                continue

            if manifest_info.content_type in (MEDIA_TYPE_OCI, MEDIA_TYPE_MANIFEST_V2):
                image = self.make_image(query, name, manifest_info)
                if image:
                    images.append(image)

        if len(images) == 0:
            return None

        image_list = {
            'Digest': list_info.digest,
            'MediaType': list_info.content_type,
            'Images': images,
        }

        if tags:
            image_list['Tags'] = tags

        return image_list

    def add_image_or_list(self, query, name, info, tags):
        if name not in self.repos:
            self.repos[name] = {
                "Name": name,
                "Images": [],
                "Lists": [],
            }

        repo = self.repos[name]

        if info.content_type in (MEDIA_TYPE_OCI,  MEDIA_TYPE_MANIFEST_V2):
            image = self.make_image(query, name, info, tags=tags)
            if image:
                repo["Images"].append(image)
        elif info.content_type in (MEDIA_TYPE_OCI_INDEX, MEDIA_TYPE_LIST_V2):
            image_list = self.make_list(query, name, info, tags=tags)
            if image_list:
                repo["Lists"].append(image_list)
        else:
            log.info("{}/{}: not an OCI image or image index".format(name, tags))

    def write(self):
        with open(self.config.output, 'w') as f:
            filtered_repos = (v for v in self.repos.values() if v['Images'] or v['Lists'])
            sorted_repos = sorted(filtered_repos, key=lambda r: r['Name'])
            for repo in sorted_repos:
                repo["Images"].sort(key=lambda x: x["Tags"])
                repo["Lists"].sort(key=lambda x: x["Tags"])

            json.dump({
                'Registry': self.config.registry_public,
                'Results': sorted_repos,
            }, f, sort_keys=True, indent=4, ensure_ascii=False)
        log.info("Wrote %s", self.config.output)


class Indexer(object):
    def __init__(self, config):
        self.conf = Config(config)

    def index(self):
        if not self.conf.indexes:
            log.warning("No indexes configured")
            return

        icon_store = None
        if self.conf.icons_dir is not None:
            if self.conf.icons_uri is None:
                raise RuntimeError("icons_dir is configured, but not icons_uri")

            icon_store = IconStore(self.conf.icons_dir, self.conf.icons_uri)

        indexes_by_registry = {}
        for index_config in self.conf.indexes:
            indexes = indexes_by_registry.setdefault(index_config.registry, [])
            indexes.append(Index(index_config, icon_store=icon_store))

        for registry, indexes in indexes_by_registry.items():
            query = RegistryQuery(registry)

            tag_patterns = set()
            for index in indexes:
                tag_patterns.update(index.config.tags)

            for repo, image, tags in query.iterate_images(tag_patterns):
                for index in indexes:
                    matches_tag = False
                    for tag in tags:
                        for tag_pattern in index.config.tags:
                            if tag_pattern.matches(tag):
                                matches_tag = True
                    if matches_tag:
                        index.add_image_or_list(query, repo, image, tags)

            for index in indexes:
                index.write()

        if icon_store is not None:
            icon_store.clean()



from setuptools import setup

setup(name='regindexer',
      version='0.1',
      description='fedmsg-hub plugin to maintain a static registry index',
      url='https://pagure.io/regindexer',
      author='Owen Taylor',
      author_email='otaylor@redhat.com',
      license='MIT',
      packages=['regindexer'],
      include_package_data=True,
      install_requires=[
          'click',
          'requests',
          'PyYAML',
      ],
      entry_points= {
          'console_scripts': [
              'regindexer=regindexer.cli:cli'
          ],
          'moksha.consumer': [
              "regindexer = regindexer.consumer:Consumer",
          ],
      })
